from cooptools.sectors.grids import RectGrid
from coopstructs.geometry import Rectangle
from typing import  Callable, List, Tuple, Dict, Any
import coopgame.pygamehelpers as help
from coopgame.gameTemplate import tt
from coopstructs.geometry.vectors.vectorN import Vector2
import coopgame.grids_graphs.draw_grid_utils as utils
from cooptools.coopEnum import CoopEnum
from cooptools.colors import Color
from coopgame.surfaceManager import SurfaceManager, SurfaceRegistryArgs
import pygame
from cooptools.decor import timer
import logging

logger = logging.getLogger('coopgame.pygridhandler')

class GridSurfaceType(CoopEnum):
    GRID_SURFACE_ID = 'GRID_SURFACE_ID'
    HIGHLIGHTS_SURFACE_ID = 'HIGHLIGHTS_SURFACE_ID'

GetMouseColorCallback = Callable[[], Color]
PosColorCallback = Callable[[], Dict[Vector2, Color]]
IdxColorCallback = Callable[[], Dict[int, Color]]
AreaRectCallback = Callable[[], Rectangle]
GridProvider = Callable[[], RectGrid]
KeyColorCallback = Callable[[], Dict[str, Color]]
HoverGridChangedCallback = Callable[[Vector2, Vector2], Any]

DEFAULT_DRAW_CONFIG = utils.GridDrawArgs(
            grid_draw_type = utils.GridDrawType.LINES,
            grid_color = Color.CHARCOAL,
            margin = 0,
            draw_scale_matrix = None,
            center_guide_color = Color.DARK_GREY,
            hover_color=Color.YELLOW,
            crosshairs_color=Color.DARK_CYAN
        )

class PyGridHandler:

    def __init__(self,
                 screen: pygame.Surface,
                 grid_provider: GridProvider,
                 get_game_area_rect_callback: AreaRectCallback,
                 get_highlights_callback: PosColorCallback = None,
                 get_outlined_callback: PosColorCallback = None,
                 get_highlighted_rows: IdxColorCallback = None,
                 get_highlighted_cols: IdxColorCallback = None,
                 get_toggled_key_colors: KeyColorCallback = None,
                 draw_config: utils.GridDrawArgs = None,
                 hover_changed_callback: HoverGridChangedCallback = None
                 ):
        super().__init__()
        self.hover_grid_pos = None

        self.parent_screen = screen
        self.grid_provider = grid_provider
        self._get_game_area_rect_callback = get_game_area_rect_callback
        self._get_highlights_callback = get_highlights_callback if get_highlights_callback else lambda: {}
        self._get_outlined_callback = get_outlined_callback if get_outlined_callback else lambda: {}
        self._get_highlighted_rows = get_highlighted_rows if get_highlighted_rows else lambda: {}
        self._get_highlighted_cols = get_highlighted_cols if get_highlighted_cols else lambda: {}
        self._get_toggled_key_colors = get_toggled_key_colors if get_toggled_key_colors else lambda: {}
        self._hover_changed_callback = hover_changed_callback

        self._draw_config: utils.GridDrawArgs = draw_config if draw_config else DEFAULT_DRAW_CONFIG

        self.surface_manager = SurfaceManager(
            surface_draw_callbacks=[
                SurfaceRegistryArgs(GridSurfaceType.GRID_SURFACE_ID.value, self._get_grid_surface),
                SurfaceRegistryArgs(GridSurfaceType.HIGHLIGHTS_SURFACE_ID.value, self._get_grid_overlay, frame_update=True),
            ]
        )

        self._mouse_hover_grid_pos = None

    @property
    def MouseHoverGridPos(self) -> Vector2:
        return self._mouse_hover_grid_pos

    def handle_hover(self):
        mouse_gp = self.mouse_grid_pos()

        if mouse_gp != self._mouse_hover_grid_pos and self._hover_changed_callback:
            self._hover_changed_callback(self._mouse_hover_grid_pos, mouse_gp)
        self._mouse_hover_grid_pos = mouse_gp

    def update(self):
        self.handle_hover()

    def invalidate(self):
        self.redraw()

    def redraw(self):
        self.surface_manager.redraw()

    def render(self,
               surface: pygame.Surface):
        self.update()
        self.surface_manager.render(surface, frame_update=True)

    def _get_grid_surface(self):
        surf = help.init_surface(self.parent_screen.get_size())
        utils.draw_to_surface(
            surface=surf,
            grid=self.grid_provider(),
            grid_draw_type=self._draw_config.grid_draw_type,
            grid_color=self._draw_config.grid_color,
            margin=self._draw_config.margin,
            draw_transform_matrix=self._draw_config.draw_scale_matrix,
            center_guide_color=self._draw_config.center_guide_color,
            toggled_key_colors=self._get_toggled_key_colors()
        )
        return surf

    def _get_grid_overlay(self):
        surf = help.init_surface(self.parent_screen.get_size())

        hover = (self.MouseHoverGridPos, self._draw_config.hover_color)
        if hover[0] == None:
            hover = None

        utils.draw_to_surface(
            surface=surf,
            grid=self.grid_provider(),
            draw_transform_matrix=self._draw_config.draw_scale_matrix,
            highlights=self._get_highlights_callback(),
            hover=hover,
            outlined_grid_cells=self._get_outlined_callback(),
            highlight_rows=self._get_highlighted_rows(),
            highlight_cols=self._get_highlighted_cols(),
            crosshairs_color=self._draw_config.crosshairs_color
        )

        return surf

    @timer(logger=logger, time_tracking_class=tt)
    def mouse_grid_pos(self) -> Vector2:
        mse = help.mouse_pos_as_vector()
        mouse_gpo = Vector2.from_tuple(utils.get_mouse_grid_pos(mouse_pos=mse,
                                                                game_area_rect= self._get_game_area_rect_callback(),
                                                                draw_transform_matrix=None,
                                                                grid=self.grid_provider()))
        return mouse_gpo


    def toggle_surface(self, gridSurfaceTypes: List[GridSurfaceType]):
        self.surface_manager.toggle_visible([x.value for x in gridSurfaceTypes])

    def show_all(self):
        self.surface_manager.show_all()

    def hide_all(self):
        self.surface_manager.hide_all()

if __name__ == "__main__":
    grid = RectGrid(10, 10)
    grid_pos = Vector2(5, 9)
    game_area_rect = Rectangle.from_tuple((0, 0, 1500, 2000))

    grid_handler = PyGridHandler()
    grid_def = grid_handler.grid_pos_definition(grid, grid_pos, game_area_rect)

    print(grid_def)
    print(grid_def.center)




from coopstructs.vectors import Vector2
from typing import Callable, Any, Optional
from cooptools.colors import Color
from coopgame.particleSystem.protocol import UpdateRenderableProtocol

vec2_callable = Callable[[], Vector2]
float_from_ms_callable = Callable[[int], float]
color_from_ms_callable = Callable[[int], Color]
color_callable = Callable[[], Color]
int_callable = Callable[[], int]
float_callable = Callable[[], float]
UpdateRenderableProtocol_provider = Callable[[Optional[UpdateRenderableProtocol]], UpdateRenderableProtocol]
phase_end_trigger = Callable[[], bool]
phase_callback = Callable[[UpdateRenderableProtocol], Any]


from coopgame.particleSystem.rendering.colorBehaviors import ColorBehaviorProtocol
from dataclasses import dataclass
from cooptools.colors import Color
from typing import Union
from coopgame.particleSystem.particle import Particle
import pygame
from cooptools.coopEnum import CoopEnum, auto

class RenderShape(CoopEnum):
    CIRCLE = auto()

class LifetimeType(CoopEnum):
    PARTICLE = auto()
    EMITTER = auto()

@dataclass(frozen=True)
class ParticleRenderer:
    colorBehavior: Union[Color, ColorBehaviorProtocol]
    shape: RenderShape = RenderShape.CIRCLE
    reverse: bool = False
    lifetime_type: LifetimeType = LifetimeType.PARTICLE

    def resolve_color(self, lifetime_ms):
        if type(self.colorBehavior) == Color:
            return self.colorBehavior.value
        else:
            return self.colorBehavior.resolve_at_t(lifetime_ms)

    def render(self, surface: pygame.Surface, particle: Particle, emitter):
        if self.lifetime_type == LifetimeType.PARTICLE:
            _color = self.resolve_color(particle.Lifetime_ms)
        elif self.lifetime_type == LifetimeType.EMITTER:
            _color = self.resolve_color(emitter.LifetimeMS)
        else:
            raise NotImplementedError(f"lifetime type {self.lifetime_type} has not been implemented for rendering")

        if self.shape == RenderShape.CIRCLE:
            pygame.draw.circle(surface=surface,
                               color=_color,
                               center=particle.pos.as_tuple(),
                               radius=particle.Size)


import pygame

from coopgame.particleSystem.emitter import Emitter
from typing import Dict

class EmitterComposite:

    def __init__(self, emitters: Dict[int, Emitter]):
        self._emitters = emitters

    def update(self, delta_ms: int):
        for component in self.Components:
            component.update(delta_ms)

    def render(self, surface: pygame.Surface):
        for component in sorted(self._emitters.values(), key=lambda x: self._emitters.keys()):
            component.render(surface)

    def n_particles(self):
        return sum(x.N_Particles for x in self._emitters.values())

    def toggle(self, value: bool):
        for x in self._emitters.values():
            x.toggle_started(value=value)

    @property
    def Components(self):
        return list(self._emitters.values())

    @property
    def N_Particles(self):
        return self.n_particles()


from coopgame.models.primitives.coopquaternion import CoopQuaternion
from typing import Tuple
import cooptools.geometry_utils.vector_utils as vec
from cooptools.common import rads_to_degrees, degree_to_rads
import cooptools.geometry_utils.circle_utils as circ

class Rotation:
    def __init__(self,
                 rads: Tuple[float, float, float] = None,
                 rotation_point: Tuple[float, float, float] = None
                 ):
        self._rotation_point = rotation_point if rotation_point else (0, 0, 0)
        self._rads = (0, 0, 0)
        self._quaternion = CoopQuaternion(axis=[0, 1, 0], angle=0)

        if rads is not None:
            self.update(rads=rads)


    @classmethod
    def from_rotation(cls, rotation):
        return Rotation(rads=rotation.Rads, rotation_point=rotation.RotationPoint)

    def __repr__(self):
        return str(self._rads)

    def update(self,
                rads: Tuple[float, float, float] = None,
                delta_rads: Tuple[float, float, float] = None,
                degrees: Tuple[float, float, float] = None,
                delta_degrees: Tuple[float, float, float] = None,
                rotation_point: Tuple[float, float, float] = None
                ):
        if rads is not None:
            self._rads = rads

        if delta_rads is not None:
            self._rads = tuple(map(lambda i, j: i + j, self._rads, delta_rads))

        # recursive call for degrees
        if degrees is not None or delta_degrees is not None:
            self.update(rads=tuple([degree_to_rads(x) for x in degrees]) if degrees else None,
                        delta_rads=tuple([degree_to_rads(x) for x in delta_degrees]) if delta_degrees else None)

        # set rotation point
        if rotation_point is not None:
            self._rotation_point = rotation_point


    @property
    def RotationPoint(self):
        return self._rotation_point

    @property
    def Rads(self):
        return self._rads

    @property
    def Degrees(self):
        return tuple(rads_to_degrees(x) for x in self._rads)

class Translation:
    def __init__(self,
                 translation_vector: Tuple[float, float, float] = None):
        self._translation_vector = (0, 0, 0)

        if translation_vector is not None:
            self.update(vector=translation_vector)

    def from_translation(self, translation):
        return Translation(translation_vector=translation.Vector)

    def __repr__(self):
        return str(self._translation_vector)

    def update(self,
               vector: Tuple[float, ...] = None,
               delta_vector: Tuple[float, ...] = None):
        if vector is not None:
            self._translation_vector = vector

        if delta_vector is not None:
            self._translation_vector = vec.add_vectors([self._translation_vector, delta_vector], allow_diff_lengths=True)

    @property
    def Vector(self):
        return self._translation_vector

class Scale:
    def __init__(self,
                 scale_vector: Tuple[float, float, float] = None,
                 scale_point: Tuple[float, float, float] = None,
                 ):
        self._scale_vector = scale_vector if scale_vector else (1, 1, 1)
        self._scale_point = scale_point if scale_point else (0, 0, 0)
        self._scale_adjustment = (0, 0, 0)

    def from_scale(self, scale):
        return Scale(
            scale_vector=scale.Vector,
            scale_point=scale.ScalePoint
        )

    def __repr__(self):
        return str(self._scale_vector)

    def update(self,
               scale_vector: Tuple[float, ...],
               scale_point: Tuple[float, ...] = None):
        adj_scale_point = vec.hadamard_product(scale_point, scale_vector) if scale_point else (0, 0, 0)
        self._scale_adjustment = vec.add_vectors(
            [vec.hadamard_product(self._scale_adjustment, scale_vector, allow_different_lengths=True), adj_scale_point,
             tuple(-1 * x for x in adj_scale_point)])
        self._scale_vector = vec.hadamard_product(self._scale_vector, scale_vector, allow_different_lengths=True)

    @property
    def ScaleAdjustment(self):
        return self._scale_adjustment


    @property
    def ScalePoint(self):
        return self._scale_point


    @property
    def Vector(self):
        return self._scale_vector


class Transform:
    def __init__(self, translation: vec.FloatVec = None,
                 rotation: vec.FloatVec = None ,
                 scale: vec.FloatVec = None):
        self._translation: Translation = Translation(translation_vector=translation)
        self._rotation: Rotation = Rotation(rads=rotation)
        self._scale: Scale = Scale(scale_vector=scale)

    @classmethod
    def from_transform(cls, transform):
        return Transform(
            translation=transform.Translation.Vector,
            scale=transform.Scale.Vector,
            rotation=transform.Rotation.Rads
        )


    @property
    def Translation(self):
        return self._translation

    @property
    def Rotation(self):
        return self._rotation

    @property
    def Scale(self):
        return self._scale

    def __repr__(self):
        return f"T{self.Translation}, R{self.Rotation}, S{self.Scale}"



from typing import Protocol
import pygame


class UpdateRenderableProtocol(Protocol):
    def update(self, delta_ms: int):
        pass

    def render(self, surf: pygame.Surface):
        pass

    def n_particles(self):
        pass



from coopgame.spriteHandling.sprites import MySprite
from coopgame.models.primitives.orientation import Transform
from typing import Dict, List, Tuple
import pygame
import coopgame.pygamehelpers as help
from coopstructs.geometry.vectors.vectorN import Vector2
from cooptools.colors import Color
import coopgame.spriteHandling.utils as s_util
from dataclasses import dataclass
import coopgame.pointdrawing.point_draw_utils as pdu

@dataclass
class SpriteTransformMapKey:
    id: str
    type: str

    def __hash__(self):
        return hash(self.id)

class SpriteTransformMapper:
    def __init__(self, sprite_transform_map: Dict[SpriteTransformMapKey, Tuple[pygame.Surface, Transform]] = None):
        self._sprite_transform_map: Dict[SpriteTransformMapKey, Tuple[pygame.Surface, Transform]] = {}

        if sprite_transform_map is not None:
            self.register(sprite_transform_map)

    def register(self, sprite_map: Dict[SpriteTransformMapKey, Tuple[pygame.Surface, Transform]]):
        self._sprite_transform_map = {**self._sprite_transform_map, **sprite_map}

    def get_surfaces(self) -> List[Tuple[pygame.Surface, pygame.Rect]]:
        return [
            s_util.transform_sprite(map[0], map[1]) for k, map in self._sprite_transform_map.items()
        ]

    def unregister(self, ids: List[SpriteTransformMapKey]):
        self._sprite_transform_map = {k: v for k, v in self._sprite_transform_map.items() if k not in ids}

    def draw(self, surface: pygame.Surface, debug: bool = False):
        sprites = self.get_surfaces()

        for sprite, pos in sprites:
            surface.blit(sprite, pos)
            if debug:
                help.draw_circle(surface, Vector2.from_tuple(pos.center), radius=3, color=Color.ORANGE, outline_color=Color.BLACK)

    @property
    def SpriteMap(self) -> Dict[SpriteTransformMapKey, Tuple[pygame.Surface, Transform]]:
        return self._sprite_transform_map

if __name__ == "__main__":
    from coopgame.spriteHandling.spritesheet import SpriteSheet
    from coopgame.handleKeyPressedArgs import InputStateHandler, InputAction, PyKeys, PyMouse, InputEvent, basic_key_actions, InputState, pygame_event_handler
    import logging
    from coopgame.gameTimeTracker import GameTimeTracker

    logging.basicConfig(level=logging.INFO)

    pygame.init()
    screen = pygame.display.set_mode((500, 500))


    mapper = SpriteTransformMapper()

    player_sprite = SpriteSheet(
        r'C:\Users\tburns\Downloads\craftpix-net-786503-free-spaceship-pixel-art-sprite-sheets\Fighter\Idle.png', 1,
        1).load_row_strip(0, colorkey=Color.WHITE.value)[0]

    transform = Transform()

    mapper.register({player_sprite: transform})

    print(mapper.get_surfaces())
    transform.Translation.update(vector=(250, 250))
    transform.Scale.update(scale_vector=(2, 2))
    transform.Rotation.update(rotation_point=(26, 32))

    input_state_handler = InputStateHandler(
        key_actions=basic_key_actions(
            up_package=lambda x: transform.Translation.update(delta_vector=(0, -1)),
            down_package=lambda x: transform.Translation.update(delta_vector=(0, 1)),
            left_package=lambda x: transform.Translation.update(delta_vector=(-1, 0)),
            right_package=lambda x: transform.Translation.update(delta_vector=(1, 0)),
            scale_up_package=lambda x: transform.Scale.update(scale_vector=(1.1, 1.1)),
            scale_down_package=lambda x: transform.Scale.update(scale_vector=(0.9, 0.9)),
            rot_left_package=lambda x: transform.Rotation.update(delta_rads=(0.005, 0)),
            rot_right_package=lambda x: transform.Rotation.update(delta_rads=(-0.005, 0)),
        ))
    game_time_tracker = GameTimeTracker(max_fps=60)
    game_time_tracker.set_start()

    while True:
        input_state = InputState(game_time_tracker.update())
        for event in pygame.event.get():
            if event.type == pygame.QUIT:
                exit()

            input_state = pygame_event_handler(input_state, event=event)

        input_state_handler.handle_input(input_state)

        # transform.Translation.update(vector=pygame.mouse.get_pos())
        # transform.Rotation.update(delta_rads=(0.0005, 0, 0))
        sprites = mapper.get_surfaces()

        screen.fill((0, 0, 0))  # fill the screen with black
        #
        # render_sprite, render_rect = rotate(player_sprite, angle=45, pivot=(player_sprite.get_size()[0] / 2, player_sprite.get_size()[1] / 2))
        # screen.blit(render_sprite, render_rect)

        for sprite, pos in sprites:
            screen.blit(sprite, pos)  # draw the ball
            help.draw_circle(screen, Vector2.from_tuple(pos.center), radius=3, color=Color.ORANGE, outline_color=Color.BLACK)
            help.draw_circle(screen, Vector2.from_tuple(transform.Translation.Vector), radius=3, color=Color.BLUE,
                             outline_color=Color.BLACK)

        pygame.display.update()  # update the screen

        # print(transform)


import pygame
from cooptools.coopEnum import CoopEnum

def _print_fonts_for_enum():
    for font in pygame.font.get_fonts():
        print(f"{font.upper()} = \"{font}\"")

class PyFont(CoopEnum):
    ARIAL = "arial"
    ARIALBLACK = "arialblack"
    BAHNSCHRIFT = "bahnschrift"
    CALIBRI = "calibri"
    CAMBRIACAMBRIAMATH = "cambriacambriamath"
    CAMBRIA = "cambria"
    CANDARA = "candara"
    COMICSANSMS = "comicsansms"
    CONSOLAS = "consolas"
    CONSTANTIA = "constantia"
    CORBEL = "corbel"
    COURIERNEW = "couriernew"
    EBRIMA = "ebrima"
    FRANKLINGOTHICMEDIUM = "franklingothicmedium"
    GABRIOLA = "gabriola"
    GADUGI = "gadugi"
    GEORGIA = "georgia"
    IMPACT = "impact"
    INKFREE = "inkfree"
    JAVANESETEXT = "javanesetext"
    LEELAWADEEUI = "leelawadeeui"
    LEELAWADEEUISEMILIGHT = "leelawadeeuisemilight"
    LUCIDACONSOLE = "lucidaconsole"
    LUCIDASANS = "lucidasans"
    MALGUNGOTHIC = "malgungothic"
    MALGUNGOTHICSEMILIGHT = "malgungothicsemilight"
    MICROSOFTHIMALAYA = "microsofthimalaya"
    MICROSOFTJHENGHEIMICROSOFTJHENGHEIUI = "microsoftjhengheimicrosoftjhengheiui"
    MICROSOFTJHENGHEIMICROSOFTJHENGHEIUIBOLD = "microsoftjhengheimicrosoftjhengheiuibold"
    MICROSOFTJHENGHEIMICROSOFTJHENGHEIUILIGHT = "microsoftjhengheimicrosoftjhengheiuilight"
    MICROSOFTNEWTAILUE = "microsoftnewtailue"
    MICROSOFTPHAGSPA = "microsoftphagspa"
    MICROSOFTSANSSERIF = "microsoftsansserif"
    MICROSOFTTAILE = "microsofttaile"
    MICROSOFTYAHEIMICROSOFTYAHEIUI = "microsoftyaheimicrosoftyaheiui"
    MICROSOFTYAHEIMICROSOFTYAHEIUIBOLD = "microsoftyaheimicrosoftyaheiuibold"
    MICROSOFTYAHEIMICROSOFTYAHEIUILIGHT = "microsoftyaheimicrosoftyaheiuilight"
    MICROSOFTYIBAITI = "microsoftyibaiti"
    MINGLIUEXTBPMINGLIUEXTBMINGLIUHKSCSEXTB = "mingliuextbpmingliuextbmingliuhkscsextb"
    MONGOLIANBAITI = "mongolianbaiti"
    MSGOTHICMSUIGOTHICMSPGOTHIC = "msgothicmsuigothicmspgothic"
    MVBOLI = "mvboli"
    MYANMARTEXT = "myanmartext"
    NIRMALAUI = "nirmalaui"
    NIRMALAUISEMILIGHT = "nirmalauisemilight"
    PALATINOLINOTYPE = "palatinolinotype"
    SEGOEMDL2ASSETS = "segoemdl2assets"
    SEGOEPRINT = "segoeprint"
    SEGOESCRIPT = "segoescript"
    SEGOEUI = "segoeui"
    SEGOEUIBLACK = "segoeuiblack"
    SEGOEUIEMOJI = "segoeuiemoji"
    SEGOEUIHISTORIC = "segoeuihistoric"
    SEGOEUISEMIBOLD = "segoeuisemibold"
    SEGOEUISEMILIGHT = "segoeuisemilight"
    SEGOEUISYMBOL = "segoeuisymbol"
    SIMSUNNSIMSUN = "simsunnsimsun"
    SIMSUNEXTB = "simsunextb"
    SITKASMALLSITKATEXTSITKASUBHEADINGSITKAHEADINGSITKADISPLAYSITKABANNER = "sitkasmallsitkatextsitkasubheadingsitkaheadingsitkadisplaysitkabanner"
    SITKASMALLSITKATEXTBOLDSITKASUBHEADINGBOLDSITKAHEADINGBOLDSITKADISPLAYBOLDSITKABANNERBOLD = "sitkasmallsitkatextboldsitkasubheadingboldsitkaheadingboldsitkadisplayboldsitkabannerbold"
    SITKASMALLSITKATEXTBOLDITALICSITKASUBHEADINGBOLDITALICSITKAHEADINGBOLDITALICSITKADISPLAYBOLDITALICSITKABANNERBOLDITALIC = "sitkasmallsitkatextbolditalicsitkasubheadingbolditalicsitkaheadingbolditalicsitkadisplaybolditalicsitkabannerbolditalic"
    SITKASMALLSITKATEXTITALICSITKASUBHEADINGITALICSITKAHEADINGITALICSITKADISPLAYITALICSITKABANNERITALIC = "sitkasmallsitkatextitalicsitkasubheadingitalicsitkaheadingitalicsitkadisplayitalicsitkabanneritalic"
    SYLFAEN = "sylfaen"
    SYMBOL = "symbol"
    TAHOMA = "tahoma"
    TIMESNEWROMAN = "timesnewroman"
    TREBUCHETMS = "trebuchetms"
    VERDANA = "verdana"
    WEBDINGS = "webdings"
    WINGDINGS = "wingdings"
    YUGOTHICYUGOTHICUISEMIBOLDYUGOTHICUIBOLD = "yugothicyugothicuisemiboldyugothicuibold"
    YUGOTHICYUGOTHICUILIGHT = "yugothicyugothicuilight"
    YUGOTHICMEDIUMYUGOTHICUIREGULAR = "yugothicmediumyugothicuiregular"
    YUGOTHICREGULARYUGOTHICUISEMILIGHT = "yugothicregularyugothicuisemilight"
    HOLOMDL2ASSETS = "holomdl2assets"
    AGENCYFB = "agencyfb"
    ALGERIAN = "algerian"
    BOOKANTIQUA = "bookantiqua"
    ARIALROUNDED = "arialrounded"
    BASKERVILLEOLDFACE = "baskervilleoldface"
    BAUHAUS93 = "bauhaus93"
    BELL = "bell"
    BERNARDCONDENSED = "bernardcondensed"
    BODONI = "bodoni"
    BODONIBLACK = "bodoniblack"
    BODONICONDENSED = "bodonicondensed"
    BODONIPOSTERCOMPRESSED = "bodonipostercompressed"
    BOOKMANOLDSTYLE = "bookmanoldstyle"
    BRADLEYHANDITC = "bradleyhanditc"
    BRITANNIC = "britannic"
    BERLINSANSFB = "berlinsansfb"
    BERLINSANSFBDEMI = "berlinsansfbdemi"
    BROADWAY = "broadway"
    BRUSHSCRIPT = "brushscript"
    BOOKSHELFSYMBOL7 = "bookshelfsymbol7"
    CALIFORNIANFB = "californianfb"
    CALISTO = "calisto"
    CASTELLAR = "castellar"
    CENTURYSCHOOLBOOK = "centuryschoolbook"
    CENTAUR = "centaur"
    CENTURY = "century"
    CHILLER = "chiller"
    COLONNA = "colonna"
    COOPERBLACK = "cooperblack"
    COPPERPLATEGOTHIC = "copperplategothic"
    CURLZ = "curlz"
    DUBAI = "dubai"
    DUBAIMEDIUM = "dubaimedium"
    DUBAIREGULAR = "dubairegular"
    ELEPHANT = "elephant"
    ENGRAVERS = "engravers"
    ERASITC = "erasitc"
    ERASDEMIITC = "erasdemiitc"
    ERASMEDIUMITC = "erasmediumitc"
    FELIXTITLING = "felixtitling"
    FORTE = "forte"
    FRANKLINGOTHICBOOK = "franklingothicbook"
    FRANKLINGOTHICDEMI = "franklingothicdemi"
    FRANKLINGOTHICDEMICOND = "franklingothicdemicond"
    FRANKLINGOTHICHEAVY = "franklingothicheavy"
    FRANKLINGOTHICMEDIUMCOND = "franklingothicmediumcond"
    FREESTYLESCRIPT = "freestylescript"
    FRENCHSCRIPT = "frenchscript"
    FOOTLIGHT = "footlight"
    GARAMOND = "garamond"
    GIGI = "gigi"
    GILLSANS = "gillsans"
    GILLSANSCONDENSED = "gillsanscondensed"
    GILLSANSULTRACONDENSED = "gillsansultracondensed"
    GILLSANSULTRA = "gillsansultra"
    GLOUCESTEREXTRACONDENSED = "gloucesterextracondensed"
    GILLSANSEXTCONDENSED = "gillsansextcondensed"
    CENTURYGOTHIC = "centurygothic"
    GOUDYOLDSTYLE = "goudyoldstyle"
    GOUDYSTOUT = "goudystout"
    HARLOWSOLID = "harlowsolid"
    HARRINGTON = "harrington"
    HAETTENSCHWEILER = "haettenschweiler"
    HIGHTOWERTEXT = "hightowertext"
    IMPRINTSHADOW = "imprintshadow"
    INFORMALROMAN = "informalroman"
    BLACKADDERITC = "blackadderitc"
    EDWARDIANSCRIPTITC = "edwardianscriptitc"
    KRISTENITC = "kristenitc"
    JOKERMAN = "jokerman"
    JUICEITC = "juiceitc"
    KUNSTLERSCRIPT = "kunstlerscript"
    WIDELATIN = "widelatin"
    LUCIDABRIGHT = "lucidabright"
    LUCIDACALLIGRAPHY = "lucidacalligraphy"
    LUCIDAFAXREGULAR = "lucidafaxregular"
    LUCIDAFAX = "lucidafax"
    LUCIDAHANDWRITING = "lucidahandwriting"
    LUCIDASANSREGULAR = "lucidasansregular"
    LUCIDASANSROMAN = "lucidasansroman"
    LUCIDASANSTYPEWRITERREGULAR = "lucidasanstypewriterregular"
    LUCIDASANSTYPEWRITER = "lucidasanstypewriter"
    LUCIDASANSTYPEWRITEROBLIQUE = "lucidasanstypewriteroblique"
    MAGNETO = "magneto"
    MAIANDRAGD = "maiandragd"
    MATURASCRIPTCAPITALS = "maturascriptcapitals"
    MISTRAL = "mistral"
    MODERNNO20 = "modernno20"
    MONOTYPECORSIVA = "monotypecorsiva"
    EXTRA = "extra"
    NIAGARAENGRAVED = "niagaraengraved"
    NIAGARASOLID = "niagarasolid"
    OCRAEXTENDED = "ocraextended"
    OLDENGLISHTEXT = "oldenglishtext"
    ONYX = "onyx"
    MSOUTLOOK = "msoutlook"
    PALACESCRIPT = "palacescript"
    PAPYRUS = "papyrus"
    PARCHMENT = "parchment"
    PERPETUA = "perpetua"
    PERPETUATITLING = "perpetuatitling"
    PLAYBILL = "playbill"
    POORRICHARD = "poorrichard"
    PRISTINA = "pristina"
    RAGE = "rage"
    RAVIE = "ravie"
    MSREFERENCESANSSERIF = "msreferencesansserif"
    MSREFERENCESPECIALTY = "msreferencespecialty"
    ROCKWELLCONDENSED = "rockwellcondensed"
    ROCKWELL = "rockwell"
    ROCKWELLEXTRA = "rockwellextra"
    SCRIPT = "script"
    SHOWCARDGOTHIC = "showcardgothic"
    SNAPITC = "snapitc"
    STENCIL = "stencil"
    TWCEN = "twcen"
    TWCENCONDENSED = "twcencondensed"
    TWCENCONDENSEDEXTRA = "twcencondensedextra"
    TEMPUSSANSITC = "tempussansitc"
    VINERHANDITC = "vinerhanditc"
    VIVALDI = "vivaldi"
    VLADIMIRSCRIPT = "vladimirscript"
    WINGDINGS2 = "wingdings2"
    WINGDINGS3 = "wingdings3"
    TEAMVIEWER15 = "teamviewer15"
    BANKGOTHIC = "bankgothic"
    BANKGOTHICMEDIUM = "bankgothicmedium"
    CITYBLUEPRINT = "cityblueprint"
    COMMERCIALPI = "commercialpi"
    COMMERCIALSCRIPT = "commercialscript"
    COUNTRYBLUEPRINT = "countryblueprint"
    DUTCH801ROMAN = "dutch801roman"
    DUTCH801 = "dutch801"
    DUTCH801EXTRA = "dutch801extra"
    EUROROMANOBLIQUE = "euroromanoblique"
    EUROROMAN = "euroroman"
    ISOCPEUR = "isocpeur"
    ISOCTEUR = "isocteur"
    MONOSPACE821 = "monospace821"
    PANROMAN = "panroman"
    ROMANTIC = "romantic"
    ROMANS = "romans"
    SANSSERIFBOLDOBLIQUE = "sansserifboldoblique"
    SANSSERIF = "sansserif"
    SANSSERIFOBLIQUE = "sansserifoblique"
    STYLUS = "stylus"
    SUPERFRENCH = "superfrench"
    SWISS721 = "swiss721"
    SWISS721OUTLINE = "swiss721outline"
    SWISS721CONDENSED = "swiss721condensed"
    SWISS721CONDENSEDOUTLINE = "swiss721condensedoutline"
    SWISS721BLACKCONDENSED = "swiss721blackcondensed"
    SWISS721EXTENDED = "swiss721extended"
    SWISS721BLACKEXTENDED = "swiss721blackextended"
    SWISS721BLACK = "swiss721black"
    SWISS721BLACKOUTLINE = "swiss721blackoutline"
    TECHNICBOLD = "technicbold"
    TECHNICLITE = "techniclite"
    TECHNIC = "technic"
    UNIVERSALMATH1 = "universalmath1"
    VINETA = "vineta"
    ACADEREF = "acaderef"
    AIGDT = "aigdt"
    AMDTSYMBOLS = "amdtsymbols"
    AMGDT = "amgdt"
    GENISO = "geniso"
    COMPLEX = "complex"
    GDT = "gdt"
    GOTHICE = "gothice"
    GOTHICG = "gothicg"
    GOTHICI = "gothici"
    GREEKC = "greekc"
    GREEKS = "greeks"
    ISOCP2 = "isocp2"
    ISOCP3 = "isocp3"
    ISOCP = "isocp"
    ISOCT2 = "isoct2"
    ISOCT3 = "isoct3"
    ISOCT = "isoct"
    ITALICC = "italicc"
    ITALICT = "italict"
    MONOTXT = "monotxt"
    PROXY1 = "proxy1"
    PROXY2 = "proxy2"
    PROXY3 = "proxy3"
    PROXY4 = "proxy4"
    PROXY5 = "proxy5"
    PROXY6 = "proxy6"
    PROXY7 = "proxy7"
    PROXY8 = "proxy8"
    PROXY9 = "proxy9"
    ROMANC = "romanc"
    ROMAND = "romand"
    ROMANT = "romant"
    SCRIPTC = "scriptc"
    SCRIPTS = "scripts"
    SIMPLEX = "simplex"
    SYASTRO = "syastro"
    SYMAP = "symap"
    SYMATH = "symath"
    SYMETEO = "symeteo"
    SYMUSIC = "symusic"
    TXT = "txt"
    LEELAWADEE = "leelawadee"
    MICROSOFTUIGHUR = "microsoftuighur"
    DOSIS = "dosis"
    DOSISEXTRABOLD = "dosisextrabold"
    DOSISEXTRALIGHT = "dosisextralight"
    DOSISMEDIUM = "dosismedium"
    DOSISREGULAR = "dosisregular"
    DOSISSEMIBOLD = "dosissemibold"
    LATOBLACK = "latoblack"
    LATO = "lato"
    LATOLIGHT = "latolight"
    LATOLIGHTITALIC = "latolightitalic"
    LATOREGULAR = "latoregular"
    LATOHAIRLINE = "latohairline"
    LATOHAIRLINEITALIC = "latohairlineitalic"
    MONTSERRATBLACK = "montserratblack"
    MONTSERRAT = "montserrat"
    MONTSERRATEXTRABOLD = "montserratextrabold"
    MONTSERRATEXTRALIGHT = "montserratextralight"
    MONTSERRATMEDIUM = "montserratmedium"
    MONTSERRATREGULAR = "montserratregular"
    MONTSERRATSEMIBOLD = "montserratsemibold"
    MONTSERRATTHIN = "montserratthin"
    OSWALD = "oswald"
    OSWALDEXTRALIGHT = "oswaldextralight"
    OSWALDMEDIUM = "oswaldmedium"
    OSWALDREGULAR = "oswaldregular"
    OSWALDSEMIBOLD = "oswaldsemibold"
    SOURCESANSPROBLACK = "sourcesansproblack"
    SOURCESANSPRO = "sourcesanspro"
    SOURCESANSPROEXTRALIGHT = "sourcesansproextralight"
    SOURCESANSPROREGULAR = "sourcesansproregular"
    SOURCESANSPROSEMIBOLD = "sourcesansprosemibold"
    BARLOWCONDENSEDBLACK = "barlowcondensedblack"
    BARLOWCONDENSED = "barlowcondensed"
    BARLOWCONDENSEDEXTRABOLD = "barlowcondensedextrabold"
    BARLOWCONDENSEDEXTRALIGHT = "barlowcondensedextralight"
    BARLOWCONDENSEDMEDIUM = "barlowcondensedmedium"
    BARLOWCONDENSEDREGULAR = "barlowcondensedregular"
    BARLOWCONDENSEDSEMIBOLD = "barlowcondensedsemibold"
    BARLOWCONDENSEDTHIN = "barlowcondensedthin"
    CAVEAT = "caveat"
    CAVEATREGULAR = "caveatregular"
    CORMORANTINFANT = "cormorantinfant"
    CORMORANTINFANTMEDIUM = "cormorantinfantmedium"
    CORMORANTINFANTREGULAR = "cormorantinfantregular"
    CORMORANTINFANTSEMIBOLD = "cormorantinfantsemibold"
    NOTOSANS = "notosans"
    NOTOSERIF = "notoserif"
    OPENSANS = "opensans"
    OPENSANSEXTRABOLD = "opensansextrabold"
    OPENSANSREGULAR = "opensansregular"
    OPENSANSSEMIBOLD = "opensanssemibold"
    RALEWAYBLACK = "ralewayblack"
    RALEWAY = "raleway"
    RALEWAYEXTRABOLD = "ralewayextrabold"
    RALEWAYEXTRALIGHT = "ralewayextralight"
    RALEWAYMEDIUM = "ralewaymedium"
    RALEWAYREGULAR = "ralewayregular"
    RALEWAYSEMIBOLD = "ralewaysemibold"
    RALEWAYTHIN = "ralewaythin"
    ROBOTOBLACK = "robotoblack"
    ROBOTO = "roboto"
    ROBOTOMEDIUM = "robotomedium"
    ROBOTOTHIN = "robotothin"
    ROBOTOCONDENSED = "robotocondensed"
    ROBOTOSLABBLACK = "robotoslabblack"
    ROBOTOSLAB = "robotoslab"
    ROBOTOSLABEXTRABOLD = "robotoslabextrabold"
    ROBOTOSLABEXTRALIGHT = "robotoslabextralight"
    ROBOTOSLABMEDIUM = "robotoslabmedium"
    ROBOTOSLABREGULAR = "robotoslabregular"
    ROBOTOSLABSEMIBOLD = "robotoslabsemibold"
    ROBOTOSLABTHIN = "robotoslabthin"
    ZILLASLAB = "zillaslab"
    ZILLASLABMEDIUM = "zillaslabmedium"
    ZILLASLABSEMIBOLD = "zillaslabsemibold"


import pygame
from cooptools.coopEnum import CoopEnum, auto

class PyKeys(CoopEnum):
    K_BACKSPACE = pygame.K_BACKSPACE
    K_TAB = pygame.K_TAB
    K_CLEAR = pygame.K_CLEAR
    K_RETURN = pygame.K_RETURN
    K_PAUSE = pygame.K_PAUSE
    K_ESCAPE = pygame.K_ESCAPE
    K_SPACE = pygame.K_SPACE
    K_EXCLAIM = pygame.K_EXCLAIM
    K_QUOTEDBL = pygame.K_QUOTEDBL
    K_HASH = pygame.K_HASH
    K_DOLLAR = pygame.K_DOLLAR
    K_AMPERSAND = pygame.K_AMPERSAND
    K_QUOTE = pygame.K_QUOTE
    K_LEFTPAREN = pygame.K_LEFTPAREN
    K_RIGHTPAREN = pygame.K_RIGHTPAREN
    K_ASTERISK = pygame.K_ASTERISK
    K_PLUS = pygame.K_PLUS
    K_COMMA = pygame.K_COMMA
    K_MINUS = pygame.K_MINUS
    K_PERIOD = pygame.K_PERIOD
    K_SLASH = pygame.K_SLASH
    K_0 = pygame.K_0
    K_1 = pygame.K_1
    K_2 = pygame.K_2
    K_3 = pygame.K_3
    K_4 = pygame.K_4
    K_5 = pygame.K_5
    K_6 = pygame.K_6
    K_7 = pygame.K_7
    K_8 = pygame.K_8
    K_9 = pygame.K_9
    K_COLON = pygame.K_COLON
    K_SEMICOLON = pygame.K_SEMICOLON
    K_LESS = pygame.K_LESS
    K_EQUALS = pygame.K_EQUALS
    K_GREATER = pygame.K_GREATER
    K_QUESTION = pygame.K_QUESTION
    K_AT = pygame.K_AT
    K_LEFTBRACKET = pygame.K_LEFTBRACKET
    K_BACKSLASH = pygame.K_BACKSLASH
    K_RIGHTBRACKET = pygame.K_RIGHTBRACKET
    K_CARET = pygame.K_CARET
    K_UNDERSCORE = pygame.K_UNDERSCORE
    K_BACKQUOTE = pygame.K_BACKQUOTE
    K_a = pygame.K_a
    K_b = pygame.K_b
    K_c = pygame.K_c
    K_d = pygame.K_d
    K_e = pygame.K_e
    K_f = pygame.K_f
    K_g = pygame.K_g
    K_h = pygame.K_h
    K_i = pygame.K_i
    K_j = pygame.K_j
    K_k = pygame.K_k
    K_l = pygame.K_l
    K_m = pygame.K_m
    K_n = pygame.K_n
    K_o = pygame.K_o
    K_p = pygame.K_p
    K_q = pygame.K_q
    K_r = pygame.K_r
    K_s = pygame.K_s
    K_t = pygame.K_t
    K_u = pygame.K_u
    K_v = pygame.K_v
    K_w = pygame.K_w
    K_x = pygame.K_x
    K_y = pygame.K_y
    K_z = pygame.K_z
    K_DELETE = pygame.K_DELETE
    K_KP0 = pygame.K_KP0
    K_KP1 = pygame.K_KP1
    K_KP2 = pygame.K_KP2
    K_KP3 = pygame.K_KP3
    K_KP4 = pygame.K_KP4
    K_KP5 = pygame.K_KP5
    K_KP6 = pygame.K_KP6
    K_KP7 = pygame.K_KP7
    K_KP8 = pygame.K_KP8
    K_KP9 = pygame.K_KP9
    K_KP_PERIOD = pygame.K_KP_PERIOD
    K_KP_DIVIDE = pygame.K_KP_DIVIDE
    K_KP_MULTIPLY = pygame.K_KP_MULTIPLY
    K_KP_MINUS = pygame.K_KP_MINUS
    K_KP_PLUS = pygame.K_KP_PLUS
    K_KP_ENTER = pygame.K_KP_ENTER
    K_KP_EQUALS = pygame.K_KP_EQUALS
    K_UP = pygame.K_UP
    K_DOWN = pygame.K_DOWN
    K_RIGHT = pygame.K_RIGHT
    K_LEFT = pygame.K_LEFT
    K_INSERT = pygame.K_INSERT
    K_HOME = pygame.K_HOME
    K_END = pygame.K_END
    K_PAGEUP = pygame.K_PAGEUP
    K_PAGEDOWN = pygame.K_PAGEDOWN
    K_F1 = pygame.K_F1
    K_F2 = pygame.K_F2
    K_F3 = pygame.K_F3
    K_F4 = pygame.K_F4
    K_F5 = pygame.K_F5
    K_F6 = pygame.K_F6
    K_F7 = pygame.K_F7
    K_F8 = pygame.K_F8
    K_F9 = pygame.K_F9
    K_F10 = pygame.K_F10
    K_F11 = pygame.K_F11
    K_F12 = pygame.K_F12
    K_F13 = pygame.K_F13
    K_F14 = pygame.K_F14
    K_F15 = pygame.K_F15
    K_NUMLOCK = pygame.K_NUMLOCK
    K_CAPSLOCK = pygame.K_CAPSLOCK
    K_SCROLLOCK = pygame.K_SCROLLOCK
    K_RSHIFT = pygame.K_RSHIFT
    K_LSHIFT = pygame.K_LSHIFT
    K_RCTRL = pygame.K_RCTRL
    K_LCTRL = pygame.K_LCTRL
    K_RALT = pygame.K_RALT
    K_LALT = pygame.K_LALT
    K_RMETA = pygame.K_RMETA
    K_LMETA = pygame.K_LMETA
    K_LSUPER = pygame.K_LSUPER
    K_RSUPER = pygame.K_RSUPER
    K_MODE = pygame.K_MODE
    K_HELP = pygame.K_HELP
    K_PRINT = pygame.K_PRINT
    K_SYSREQ = pygame.K_SYSREQ
    K_BREAK = pygame.K_BREAK
    K_MENU = pygame.K_MENU
    K_POWER = pygame.K_POWER
    K_EURO = pygame.K_EURO

class PyMouse(CoopEnum):
    LEFT = 1
    MIDDLE = 2
    RIGHT = 3
    SCROLLUP = 4
    SCROLLDOWN = 5
    POSITION = 6

class InputType(CoopEnum):
    MOUSE=auto()
    KEYBOARD=auto()

ALL_INPUT = tuple((InputType.MOUSE, x) for x in PyMouse) + tuple((InputType.KEYBOARD, x) for x in PyKeys)

def pygame_key_mapper(val: int):
    return PyKeys.by_val(val).name






from coopgame.renderedObjectHandling.objectOrientationArgs import ObjectOrientationArgs
import math

class RenderedViewArgs:
    def __init__(self,
                 near_plane_dist: float = None,
                 far_plane_dist: float = None,
                 field_of_view_rads: float = None,
                 camera_orientation_args: ObjectOrientationArgs = None
                 ):
        self.camera_orientation = camera_orientation_args if camera_orientation_args else ObjectOrientationArgs()
        self.near_plane_dist = near_plane_dist if near_plane_dist else .1
        self.far_plane_dist = far_plane_dist if far_plane_dist else 100
        self.field_of_view_rads = field_of_view_rads if field_of_view_rads else math.pi / 2


import os
import pygame
import coopgame.spriteHandling.utils as utils

class SpriteFolderHandler:

    def __init__(self, dir, delim: str = "__"):
        if not pygame.get_init():
            pygame.init()
            pygame.display.set_mode((800, 600))

        self.poses = {}
        files = os.listdir(dir)

        for file in files:
            name_attrs = file.split(delim)
            if len(name_attrs) > 2:
                pose = '_'.join(name_attrs[0:2])
            else:
                pose = f"{name_attrs[0]}"

            self.poses.setdefault(pose, [])
            self.poses[pose].append(utils.try_load_sprite_file(filepath=f"{dir}\{file}"))

if __name__ == "__main__":
    from pprint import pprint
    sfh = SpriteFolderHandler(r'C:\Users\tburns\Downloads\ninjaadventurenew\png')
    pprint(sfh.poses)


import pygame
from coopgame.colors import Color
from typing import List, Dict, Callable
from coopstructs.geometry.curves.curves import Curve
from coopstructs.geometry import Line, PolygonRegion
import numpy as np
from shapely.geometry import LineString
import coopgame.pygamehelpers as help
from dataclasses import dataclass
import coopgame.linedrawing.line_draw_utils as lutils
import coopgame.pointdrawing.point_draw_utils as putils

@dataclass(frozen=True)
class CurveDrawArgs:
    line_args: lutils.DrawLineArgs = None
    control_point_args: putils.DrawPointArgs = None
    control_line_args: lutils.DrawLineArgs = None
    draw_scale_matrix: np.ndarray = None
    buffer: int = None
    buffer_color: Color = None
    resolution: int = 30

    @property
    def BaseArgs(self):
        return CurveDrawArgs(
            line_args=self.line_args.BaseArgs,
            draw_scale_matrix=self.draw_scale_matrix,
            resolution=self.resolution
        )

    @property
    def OverlayArgs(self):
        return CurveDrawArgs(
            draw_scale_matrix=self.draw_scale_matrix,
            control_point_args=self.control_point_args,
            control_line_args=self.control_line_args
        )

    @property
    def BufferArgs(self):
        return CurveDrawArgs(
            draw_scale_matrix=self.draw_scale_matrix,
            buffer=self.buffer,
            buffer_color=self.buffer_color,
        )

def draw_curves(curves: Dict[Curve, CurveDrawArgs],
                surface: pygame.Surface):

    for curve, args in curves.items():
        line_rep = curve.line_representation(resolution=args.resolution)

        if args.buffer_color:
            line = LineString([
                line_rep[0].origin] +
                [x.destination for x in line_rep]
            )

            dilated = line.buffer(args.buffer)
            poly = PolygonRegion.from_shapely_polygon(dilated)
            buffer_color = Color.GREEN if args.buffer_color is None else args.buffer_color
            help.draw_polygon(surface, [x.as_tuple() for x in poly.boundary_points], buffer_color)

        if args.control_line_args:
            lutils.draw_lines(
                {x: args.control_line_args for x in curve.ControlLines},
                surface=surface,
                draw_transform_matrix=args.draw_scale_matrix
            )

        lutils.draw_lines(
            lines={x: args.line_args for x in line_rep},
            surface=surface,
            draw_transform_matrix=args.draw_scale_matrix

        )

        if args.control_point_args:
            putils.draw_points(
                points={
                    x.as_tuple(): args.control_point_args for x in curve.ControlPoints
                },
                surface=surface,
                draw_scale_matrix=args.draw_scale_matrix
            )


import logging
from typing import List
from cooptools.decor import TimeableProtocol

class MonitoredClassLogger():

    def __init__(self,
                 monitored_classes: List[TimeableProtocol] = None):
        self.time_since_report_stats = 0
        self.monitored_classes: List[TimeableProtocol] = monitored_classes if monitored_classes else []


    def check_and_log(self,
                      logger: logging.Logger,
                      delta_time_ms: int,
                      log_interval_ms: int):
        self.time_since_report_stats += delta_time_ms
        if self.time_since_report_stats > log_interval_ms:
            for item in self.monitored_classes:
                print_dict = {method: round(time, 3) for method, time in item.internally_tracked_times.items()}
                logger.info(f"{item} -- {print_dict}")

            self.time_since_report_stats = 0

    def register_classes(self, new: List[TimeableProtocol]):
        self.monitored_classes += new


import time


class Status:

    def __init__(self,
                 name,
                 applied_time: float = None,
                 clear_time: float = None):
        self.name = name
        self.applied_time = applied_time if applied_time else time.clock()
        self.clear_time = clear_time





import time

from cooptools.sectors.grids import RectGrid
from cooptools.coopEnum import CardinalPosition
from coopstructs.geometry import Rectangle
import pygame
from coopgame.colors import Color
from typing import Dict, Tuple, List, Iterable, Callable
from coopgame.enums import GridDrawType
import coopgame.pygamehelpers as help
from coopstructs.geometry.vectors.vectorN import Vector2
from cooptools.sectors.rect_utils import sector_from_coord
from dataclasses import dataclass
import numpy as np
import coopgame.linedrawing.line_draw_utils as ldu
from coopstructs.geometry.lines import Line
from coopgame.gameTemplate import tt
from cooptools.decor import timer
import logging
import cooptools.matrixManipulation as mm

logger = logging.getLogger('coopgame.draw_grid_utils')

@dataclass(frozen=True)
class GridDrawArgs:
    grid_draw_type: GridDrawType = None
    grid_color: Color = None
    margin: int = 0
    draw_scale_matrix: mm.TransformMatrixProvider = None
    center_guide_color: Color = None
    hover_color: Color = None
    crosshairs_color: Color = None

def draw_to_surface(
        surface: pygame.Surface,
        grid: RectGrid,
        grid_draw_type: GridDrawType = None,
        grid_color: Color = None,
        margin: int = 0,
        draw_transform_matrix: np.array = None,
        center_guide_color: Color = None,
        highlights: Dict[Vector2, Color] = None,
        hover: Tuple[Vector2, Color] = None,
        toggled_key_colors: Dict[str, Color] = None,
        outlined_grid_cells: Dict[Vector2, Color] = None,
        highlight_rows: Dict[int, Color] = None,
        highlight_cols: Dict[int, Color] = None,
        crosshairs_color: Color = None

):
    if grid is None:
        return

    if grid_color is not None:
        _draw_grid(
            grid=grid,
            surf=surface,
            grid_draw_type=grid_draw_type,
            grid_color=grid_color,
            margin=margin,
            draw_transform_matrix=draw_transform_matrix
        )


    highlights = _resolve_highlights(
        grid=grid,
        toggled_key_colors=toggled_key_colors,
        hover=hover,
        rows=highlight_rows,
        cols=highlight_cols,
        center_guide_color=center_guide_color,
        crosshairs_color=crosshairs_color,
        additional_highlight_grid_cells=highlights
    )

    grid_box_rect = grid_box_rectangle_at_pos(surface.get_width(), surface.get_height(), grid, margin)
    _draw_highlighted_grids(surface,
                            grid=grid,
                            draw_scale_matrix=draw_transform_matrix,
                            highlights=highlights)


    _draw_outlined_grids(surface=surface
                         , grid_box_rect=grid_box_rect
                         , outlined_grid_cells=outlined_grid_cells
                         , margin=margin
                         , draw_transform_matrix=draw_transform_matrix)


@timer(logger=logger, time_tracking_class=tt)
def _draw_grid(
        grid: RectGrid,
        surf: pygame.Surface,
        grid_draw_type: GridDrawType = None,
        grid_color: Color = None,
        margin: int = 1,
        draw_transform_matrix: mm.TransformMatrixProvider= None
):
    if grid_draw_type is None:
        grid_draw_type = GridDrawType.LINES

    if grid_color is None:
        grid_color = Color.DARK_GREY

    if grid_draw_type == GridDrawType.BOXES:
        _draw_grid_boxes(surface=surf, grid=grid, margin=margin, grid_color=grid_color, transform_matrix=draw_transform_matrix)
    elif grid_draw_type == GridDrawType.LINES:
        _draw_grid_lines(surface=surf, grid=grid, margin=margin, grid_color=grid_color, draw_transform_matrix=draw_transform_matrix)

@timer(logger=logger, time_tracking_class=tt)
def _resolve_highlights(grid: RectGrid,
                        toggled_key_colors: Dict[str, Color] = None,
                        hover: Tuple[Vector2, Color] = None,
                        rows: Dict[int, Color] = None,
                        cols: Dict[int, Color] = None,
                        center_guide_color: Color = None,
                        crosshairs_color: Color = None,
                        additional_highlight_grid_cells: Dict[Vector2, Color] = None
                        ):
    ret = {}

    if hover is not None and crosshairs_color is not None:
        rows[hover[0].x] = crosshairs_color
        cols[hover[0].y] = crosshairs_color

    highligted_rc = _rc_highlight_definition(
        grid=grid,
        rows=rows,
        cols=cols,
        center_guide_color=center_guide_color,
    )
    ret = {**ret, **highligted_rc}

    if additional_highlight_grid_cells:
        ret = {**ret, **additional_highlight_grid_cells}

    if toggled_key_colors:
        for k, color in toggled_key_colors.items():
            toggled_highlights = _toggled_highlighted_grids(
                grid=grid,
                toggle_key=k,
                color=color
            )

            ret = {**ret, **toggled_highlights}

    if hover:
        ret[hover[0]] = hover[1]

    return ret

@timer(logger=logger, time_tracking_class=tt)
def _rc_highlight_definition(
        grid: RectGrid,
        rows: Dict[int, Color] = None,
        cols: Dict[int, Color] = None,
        center_guide_color: Color = None,
) -> Dict[Vector2, Color]:
    highlights = {}

    if rows is None:
        rows = {}
    if cols is None:
        cols = {}
    if center_guide_color:
        rs, cs = define_center_grids(grid, color=center_guide_color)
        rows = {**rows, **rs}
        cols = {**cols, **cs}

    row_highlights = {Vector2(r, c): color for r, color in rows.items() for c in range(0, grid.nColumns)}
    col_highlights = {Vector2(r, c): color for r in range(0, grid.nRows) for c, color in cols.items()}
    highlights.update(col_highlights)
    highlights.update(row_highlights)
    return highlights

@timer(logger=logger, time_tracking_class=tt)
def define_center_grids(grid: RectGrid,
                        color: Color) -> Tuple[Dict[int, Color], Dict[int, Color]]:
    if grid.nRows % 2 == 0:
        center_rows = [grid.nRows // 2 - 1, grid.nRows // 2]
    else:
        center_rows = [((grid.nRows - 1) // 2)]

    if grid.nColumns % 2 == 0:
        center_cols = [grid.nColumns // 2 - 1, grid.nColumns // 2]
    else:
        center_cols = [((grid.nColumns - 1) // 2)]

    return ({x: color for x in center_rows}, {x: color for x in center_cols})


# @timer
def _draw_outlined_grids(surface: pygame.Surface,
                         grid_box_rect: Rectangle,
                         margin=0,
                         outlined_grid_cells: Dict[Vector2, Color] = None,
                         draw_transform_matrix=None):

    if draw_transform_matrix is not None:
        raise ValueError("Draw grid outlines does not support a transform")

    if outlined_grid_cells is None or len(outlined_grid_cells) == 0:
        return

    for grid_pos, color in outlined_grid_cells.items():
        if color is None:
            color = Color.YELLOW

        rect = Rectangle.from_tuple(((margin + grid_box_rect.width) * grid_pos.x + margin
                                     , (margin + grid_box_rect.height) * grid_pos.y + margin
                                     , grid_box_rect.height
                                     , grid_box_rect.width))
        my_image = pygame.Surface((rect.width, rect.height), pygame.SRCALPHA)
        pygame.draw.rect(my_image, color.value, my_image.get_rect(), 3)
        surface.blit(my_image, (rect.x, rect.y))

@timer(logger=logger, time_tracking_class=tt)
def _toggled_highlighted_grids(grid: RectGrid,
                              toggle_key: str,
                              color: Color) -> Dict[Vector2, Color]:
    highlights = {}

    if toggle_key in grid.GridKeys:
        grid_value_array = grid.state_value_as_array(key=toggle_key)
        for x in range(0, grid.nColumns):
            for y in range(0, grid.nRows):
                if grid_value_array[y][x].value:
                    highlights[Vector2(y, x)] = color

    return highlights

@timer(logger=logger, time_tracking_class=tt)
def _scaled_rect_points(rects: List[Rectangle],
                        transform_matrix=None):
    points = []
    for rect in rects:
        [points.append(x) for x in rect.Corners]
    scaled = mm.point_transform_3d(points,
                                lh_matrix=transform_matrix)

    ret = []
    for ii in range(0, len(scaled), 4):
        ret.append(scaled[ii: ii+4])

    return ret

@timer(logger=logger, time_tracking_class=tt)
def _rect_at_pos(
    surface: pygame.Surface,
    grid: RectGrid,
    highlights: Dict[Vector2, Color],
    margin: int = 0
):
    pos_rects = [
        (grid_pos, grid_box_rectangle_at_pos(
                width=surface.get_width(),
                height=surface.get_height(),
                grid=grid,
                margin=margin,
                grid_pos=grid_pos))
        for grid_pos in highlights.keys()
    ]
    return pos_rects


@timer(logger=logger, time_tracking_class=tt)
def _draw_highlighted_grids(surface: pygame.Surface,
                            grid: RectGrid,
                            highlights: Dict[Vector2, Color],
                            draw_scale_matrix=None,
                            margin: int = 0):
    if draw_scale_matrix is None:
        draw_scale_matrix = np.identity(4)

    pos_rects = _rect_at_pos(
        surface=surface,
        grid=grid,
        highlights=highlights,
        margin=margin
    )
    # pos_rects = [
    #     (grid_pos, grid_box_rectangle_at_pos(
    #             width=surface.get_width(),
    #             height=surface.get_height(),
    #             grid=grid,
    #             margin=margin,
    #             grid_pos=grid_pos))
    #     for grid_pos in highlights.keys()
    # ]

    scaled_rects = _scaled_rect_points(
        rects=[x[1] for x in pos_rects],
        transform_matrix=draw_scale_matrix
    )

    _draw_plygons(
        pos_rects,
        surface,
        scaled_rects,
        highlights
    )
    # for ii, x in enumerate(pos_rects):
    #     help.draw_polygon(surface, scaled_rects[ii], highlights[x[0]])

@timer(logger=logger, time_tracking_class=tt)
def _draw_plygons(pos_rects,
                  surface,
                  scaled_rects,
                  highlights):
    for ii, x in enumerate(pos_rects):
        help.draw_polygon(surface, scaled_rects[ii], highlights[x[0]])


# @timer
def _draw_grid_boxes(surface: pygame.Surface,
                     grid: RectGrid,
                     margin=0,
                     grid_color: Color = None,
                     transform_matrix=None):

    if transform_matrix is None:
        transform_matrix = np.identity(4)

    for y in range(0, grid.nRows):
        for x in range(0, grid.nColumns):
            grid_box_rect = grid_box_rectangle_at_pos(
                width=surface.get_width(),
                height=surface.get_height(),
                grid=grid,
                grid_pos=Vector2(x, y),
                margin=margin)
            points = help.scaled_points_of_a_rect(grid_box_rect, transform_matrix)[:, :2]
            help.draw_polygon(surface, list(points), grid_color)


# @timer
def _draw_grid_lines(surface: pygame.Surface,
                     grid: RectGrid,
                     margin=0,
                     grid_color: Color = None,
                     draw_transform_matrix: mm.TransformMatrixProvider =None):
    grid_box_rect = grid_box_rectangle_at_pos(
        width=surface.get_width(),
        height=surface.get_height(),
        grid=grid,
        grid_pos=Vector2(0, 0),
        margin=margin)

    if draw_transform_matrix is None:
        draw_transform_matrix = np.identity(4)

    '''Draw each line'''
    lines = []

    for x in range(0, grid.nColumns + 1):
        point_x1 = x * (grid_box_rect.width + margin)
        point_y1 = 0
        point_x2 = x * (grid_box_rect.width + margin)
        point_y2 = (grid_box_rect.height + margin) * grid.nRows
        lines.append(Line(
            origin=(point_x1, point_y1),
            destination=(point_x2, point_y2),
        ))
    for y in range(0, grid.nRows + 1):
        s = (0, y * (grid_box_rect.height + margin))
        e = ((grid_box_rect.width + margin) * grid.nColumns, y * (grid_box_rect.height + margin))
        lines.append(Line(
            origin=s,
            destination=e
        ))

    ldu.draw_lines(
        lines={x: ldu.DrawLineArgs(
            color=grid_color) for x in lines},
        surface=surface,
        draw_transform_matrix=draw_transform_matrix
    )

# @timer
def get_mouse_grid_pos(grid: RectGrid,
                       game_area_rect: Rectangle,
                       mouse_pos: Vector2,
                       draw_transform_matrix=None):
    if draw_transform_matrix is None:
        draw_transform_matrix = np.identity(4)

    draw_scale_matrix_inv = np.linalg.inv(draw_transform_matrix)

    """Gets the mouse position and converts it to a grid position"""
    mouse_game_area_coord = help.game_area_coords_from_parent_coords(parent_coords=mouse_pos, game_area_surface_rectangle=game_area_rect)
    mouse_over_grid_coords = help.viewport_point_on_plane(mouse_game_area_coord, game_area_rect, draw_transform_matrix, margin=0)
    points = np.array([mouse_over_grid_coords[0], mouse_over_grid_coords[1], mouse_over_grid_coords[2], 1])
    grid_coords = draw_scale_matrix_inv.dot(np.transpose(points))


    # return coord_to_grid_converter(mouse_game_area_coord, game_area_rect)
    grid_pos = sector_from_coord(coord=(grid_coords[0], grid_coords[1]),
                                 sector_def=(grid.nRows, grid.nColumns),
                                 area_dims=game_area_rect.as_tuple()[2:4]
                                 )

    return grid_pos
    # return coord_to_grid_converter((grid_coords[0], grid_coords[1]), game_area_rect.Dims)


# @timer
def grid_point_to_viewport_point(surface: pygame.Surface,
                                 grid,
                                 grid_pos: Vector2,
                                 grid_point_type: CardinalPosition = CardinalPosition.CENTER,
                                 margin: int = 0,
                                 draw_transform_matrix=None):
    grid_box_rect = grid_box_rectangle_at_pos(
        width=surface.get_width(),
        height=surface.get_height(),
        grid=grid,
        margin=margin,
        grid_pos=grid_pos
    )

    grid_box_scaled_points = help.scaled_points_of_a_rect(grid_box_rect, draw_scale_matrix=draw_transform_matrix)

    if grid_point_type == CardinalPosition.CENTER:
        x = sum(point[0] for point in grid_box_scaled_points) / len(grid_box_scaled_points)
        y = sum(point[1] for point in grid_box_scaled_points) / len(grid_box_scaled_points)
        return Vector2(x, y)
    elif grid_point_type == CardinalPosition.ORIGIN:
        return Vector2(grid_box_scaled_points[0][0], grid_box_scaled_points[0][1])
    else:
        raise NotImplementedError(f"Unimplemented grid_point_type {grid_point_type}")

def grid_box_rectangle_at_pos(width,
                              height,
                              grid: RectGrid,
                              margin: int = 0,
                              grid_pos: Vector2 = None
                              ) -> Rectangle:
    if grid_pos is None:
        grid_pos = Vector2(0, 0)

    grid_box_height = height / grid.nRows
    grid_box_width = width / grid.nColumns

    # get basic grid pos rect at grid_pos
    rect = Rectangle.from_tuple((grid_pos.y * grid_box_width,
                                 grid_pos.x * grid_box_height,
                                 grid_box_width,
                                 grid_box_height))

    # add the margin
    rect = rect.with_margin(margin=margin)


    return rect

if __name__ == "__main__":

    grid = RectGrid(100, 100)
    rect = Rectangle.from_tuple((0, 0, 500, 1000))
    n=0
    t0 = time.perf_counter()
    while True:
        get_mouse_grid_pos(
            grid=grid,
            game_area_rect=rect,
            mouse_pos=Vector2(10, 10),
            draw_transform_matrix=None
        )
        n+=1
        print(n)
        t1 = time.perf_counter()
        if t1-t0 > 10:
            break


import time

class Health:

    def __init__(self,
                 max_health: float,
                 current_health: float = None):
        self._max_health = max_health
        self._current_health = current_health if current_health is not None else self._max_health
        self._death_timer = None

    def change_max_health(self, delta):
        self._max_health += delta

    def change_current_health(self, delta, game_perf: int = None):
        if self._current_health + delta <= 0:
            self._current_health = 0
            self._death_timer = time.perf_counter() if game_perf is None else game_perf
        elif self._current_health + delta > self._max_health:
            self._current_health = self._max_health
        else:
            self._current_health += delta

        return self._current_health

    def set_full(self):
        self._current_health = self._max_health

    @property
    def Alive(self):
        return self._current_health > 0

    @property
    def Remaining(self):
        return self._current_health

    @property
    def Max(self):
        return self._max_health

    @property
    def LastDeathTime(self):
        return self._death_timer


import struct
import imghdr
from coopstructs.geometry.vectors.vectorN import Vector2
import ntpath
import os
from coopgame.models.primitives.orientation import Transform
from typing import List, Tuple
import pygame
import cooptools.geometry_utils.vector_utils as vec

def get_image_size(fname):
    '''Determine the image type of fhandle and return its size.
    from draco'''
    with open(fname, 'rb') as fhandle:
        head = fhandle.read(24)
        if len(head) != 24:
            return
        if imghdr.what(fname) == 'png':
            check = struct.unpack('>i', head[4:8])[0]
            if check != 0x0d0a1a0a:
                return
            width, height = struct.unpack('>ii', head[16:24])
        elif imghdr.what(fname) == 'gif':
            width, height = struct.unpack('<HH', head[6:10])
        elif imghdr.what(fname) == 'jpeg':
            try:
                fhandle.seek(0)  # Read 0xff next
                size = 2
                ftype = 0
                while not 0xc0 <= ftype <= 0xcf:
                    fhandle.seek(size, 1)
                    byte = fhandle.read(1)
                    while ord(byte) == 0xff:
                        byte = fhandle.read(1)
                    ftype = ord(byte)
                    size = struct.unpack('>H', fhandle.read(2))[0] - 2
                # We are at a SOFn block
                fhandle.seek(1, 1)  # Skip `precision' byte.
                height, width = struct.unpack('>HH', fhandle.read(4))
            except Exception:  # IGNORE:W0703
                return
        else:
            return
        return width, height

def image_at_rect(surf: pygame.Surface, rectangle, colorkey = None):
    """Load a specific image from a specific rectangle."""
    # Loads image from x, y, x+offset, y+offset.
    rect = pygame.Rect(rectangle)
    image = pygame.Surface(rect.size).convert()
    image.blit(surf, (0, 0), rect)
    if colorkey is not None:
        if colorkey is -1:
            colorkey = image.get_at((0,0))
        image.set_colorkey(colorkey, pygame.RLEACCEL)
    return image


def image_at(surf: pygame.Surface, pos: Vector2, pixel_dims: Tuple[int, int], colorkey = None):
    """Load a specific image from a specific rectangle."""
    # Loads image from x, y, x+offset, y+offset.
    pixel_w, pixel_h = pixel_dims
    rectangle = (pos.x * pixel_w, pos.y * pixel_h, pixel_w, pixel_h)
    return image_at_rect(surf=surf, rectangle=rectangle, colorkey=colorkey)

def images_at(rects, colorkey = None):
    """Load a whole bunch of images and return them as a list."""
    return [image_at_rect(rect, colorkey) for rect in rects]

def try_load_sprite_file(filepath):
    try:
        return pygame.image.load(filepath).convert()
    except pygame.error as e:
        print(f"Unable to load spritesheet image: {filepath}: {e}")
        raise SystemExit(e)

def files_at_dir(directory):
    # iterate over files in
    # that directory
    ret = []
    for filename in os.scandir(directory):
        if filename.is_file():
            ret.append(rf"{directory}\{filename.name}")
    return ret

def path_and_file(filepath: str):
    head, tail = ntpath.split(filepath)
    tail = tail or ntpath.basename(head)
    return head, tail


def rename_files(filepaths: List[str], replace: List[Tuple[str, str]] = None):
    for path in filepaths:
        src = path
        path, filename = path_and_file(path)

        new = filename
        for rep in replace or []:
            new = new.replace(rep[0], rep[1])

        os.rename(src=src, dst=fr"{path}\{new}")

def scaled_rotation_point(transform: Transform) -> Tuple[float, ...]:
    return vec.hadamard_product(transform.Scale.Vector[0:2], transform.Rotation.RotationPoint[0:2])

def transform_sprite(surf: pygame.Surface, transform: Transform) -> Tuple[pygame.Surface, pygame.Rect]:
    # scale
    sprite_dims = surf.get_size()
    scale_vec = vec.hadamard_product(sprite_dims, transform.Scale.Vector[0:2])
    ret = pygame.transform.scale(surf, scale_vec)

    # rotate
    scaled_rot_point = scaled_rotation_point(transform)
    ret, rect = rotate_surface(surface=ret, angle=transform.Rotation.Degrees[0], pivot=scaled_rot_point)

    # translate
    rect.center = vec.add_vectors([rect.center, transform.Translation.Vector[0:2], tuple(-x for x in scaled_rot_point)])

    return (ret, rect)

def rotate_surface(surface, angle, pivot):
    """Rotate the surface around the pivot point.

    Args:
        surface (pygame.Surface): The surface that is to be rotated.
        angle (float): Rotate by this angle.
        pivot (tuple, list, pygame.math.Vector2): The pivot point.
        offset (pygame.math.Vector2): This vector is added to the pivot.
    """
    start_rect = surface.get_rect()

    c_w = start_rect.width // 2
    if start_rect.width % 2 != 0: c_w += 1
    c_h = start_rect.height // 2
    if start_rect.height % 2 != 0: c_h += 1

    c = pygame.math.Vector2(c_w, c_h)
    # c = pygame.math.Vector2(*start_rect.center)

    rotateable_vector = c - pygame.math.Vector2(*pivot)
    rotated_vector = rotateable_vector.rotate(angle)
    rotated_surface = pygame.transform.rotate(surface, angle)
    rotated_surface_rect = rotated_surface.get_rect()

    relocation_vector = rotated_vector - rotateable_vector
    rotated_surface_rect.center = (c.x + relocation_vector.x, c.y - relocation_vector.y)

    return rotated_surface, rotated_surface_rect